#ifndef MixedCustom_HPP
#define MixedCustom_HPP

#include <QObject>

// Object source includes
// - externally generated header moc file
// - AUTOMOC generated source .moc file
class MixedCustom : public QObject
{
  Q_OBJECT
public:
  MixedCustom();
  ~MixedCustom();
};

// Function forward declaration
void moc_MixedCustom(MixedCustom const& arg);

#endif
