/*
  Copyright Kitware, Inc.
  Distributed under the OSI-approved BSD 3-Clause License.
  See accompanying file Copyright.txt for details.
*/
#ifndef KWIML_TEST_H
#define KWIML_TEST_H

/* Quiet MS standard library deprecation warnings.  */
#ifndef _CRT_SECURE_NO_DEPRECATE
# define _CRT_SECURE_NO_DEPRECATE
#endif

#else
# error "test.h included multiple times."
#endif
