#ifndef InIncludes_SubOwnDot_P_HPP
#define InIncludes_SubOwnDot_P_HPP

#include <QObject>

namespace InIncludes {

class SubOwnDotPrivate : public QObject
{
  Q_OBJECT
public:
  SubOwnDotPrivate();
  ~SubOwnDotPrivate();
};

} // End of namespace

#endif
