#ifndef OwnUnderscore_P_HPP
#define OwnUnderscore_P_HPP

#include <QObject>

class OwnUnderscorePrivate : public QObject
{
  Q_OBJECT
public:
  OwnUnderscorePrivate();
  ~OwnUnderscorePrivate();
};

#endif
