using namespace System;

namespace CSharpLibrary
{
    public ref class UsefulManagedCppClass
    {
    public:

        UsefulManagedCppClass();
        void RunTest();

    private:

        String^ m_usefulString;
    };
}
