#ifndef STYLEE_HPP
#define STYLEE_HPP

// The included file is not in the sources list and won't be detected by
// AUTOMOC source file with the same base name.
// It is registered to AUTOMOCed via a moc_<NAME>.cpp include in StyleE.cpp
// though.
#include "StyleEInclude.hpp"

#endif
