#ifndef STYLEA_HPP
#define STYLEA_HPP

#include <QObject>

/* clang-format off */
/// Q_OBJECT on a single new line
///
class StyleA : public QObject
{
  Q_OBJECT
public:
  StyleA();
};
/* clang-format on */

#endif
